"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

const serializeError = error => error != null ? {
  name: error.name == null ? null : String(error.name),
  code: error.code == null ? null : String(error.code),
  message: String(error.message),
  stack: String(error.stack)
} : null;

const buildInit = async (currentPool, basePool, readModelName, store, modelInterop, next, eventstoreAdapter) => {
  const pool = { ...basePool,
    ...currentPool
  };
  const {
    PassthroughError,
    inlineLedgerRunQuery,
    generateGuid,
    escapeStr,
    databaseNameAsId,
    ledgerTableNameAsId,
    xaKey
  } = pool;
  const rootSavePointId = generateGuid(xaKey, 'ROOT');
  await inlineLedgerRunQuery(`BEGIN TRANSACTION;
     SET TRANSACTION ISOLATION LEVEL REPEATABLE READ;
     SAVEPOINT ${rootSavePointId};
     WITH "CTE" AS (
      SELECT "XaKey" FROM ${databaseNameAsId}.${ledgerTableNameAsId}
        WHERE "EventSubscriber" = ${escapeStr(readModelName)}
        AND "XaKey" = ${escapeStr(xaKey)}
        AND "IsPaused" = FALSE
        AND "Errors" IS NULL
        FOR NO KEY UPDATE NOWAIT
    )
      SELECT 1/Count("CTE"."XaKey") AS "NonZero" FROM "CTE";
    `, true);
  const nextCursor = await eventstoreAdapter.getNextCursor(null, []);

  try {
    const handler = await modelInterop.acquireInitHandler(store);

    if (handler != null) {
      await handler();
    }

    await inlineLedgerRunQuery(`UPDATE ${databaseNameAsId}.${ledgerTableNameAsId}
       SET "SuccessEvent" = ${escapeStr(JSON.stringify({
      type: 'Init'
    }))},
       "Cursor" = ${escapeStr(JSON.stringify(nextCursor))}
       WHERE "EventSubscriber" = ${escapeStr(readModelName)};

       COMMIT;
      `);
    await next();
  } catch (error) {
    if (error instanceof PassthroughError) {
      throw error;
    }

    await inlineLedgerRunQuery(`UPDATE ${databaseNameAsId}.${ledgerTableNameAsId}
       SET "Errors" = jsonb_insert(
         COALESCE("Errors", jsonb('[]')),
         CAST(('{' || jsonb_array_length(COALESCE("Errors", jsonb('[]'))) || '}') AS TEXT[]),
         jsonb(${escapeStr(JSON.stringify(serializeError(error)))})
       ),
       "FailedEvent" = ${escapeStr(JSON.stringify({
      type: 'Init'
    }))},
       "Cursor" = ${escapeStr(JSON.stringify(nextCursor))}
       WHERE "EventSubscriber" = ${escapeStr(readModelName)};

       COMMIT;
      `);
  }
};

const buildEvents = async (currentPool, basePool, readModelName, store, modelInterop, next, eventstoreAdapter, getVacantTimeInMillis) => {
  const pool = { ...basePool,
    ...currentPool
  };
  const {
    PassthroughError,
    inlineLedgerRunQuery,
    generateGuid,
    escapeStr,
    databaseNameAsId,
    ledgerTableNameAsId,
    inputCursor,
    eventTypes,
    xaKey
  } = pool;
  let lastSuccessEvent = null;
  let lastFailedEvent = null;
  let lastError = null;
  let localContinue = true;
  let cursor = inputCursor;
  let eventsPromise = eventstoreAdapter.loadEvents({
    eventTypes,
    eventsSizeLimit: 6553600,
    limit: 100,
    cursor
  }).then(result => result != null ? result.events : []);
  let rootSavePointId = generateGuid(xaKey, 'ROOT');
  await inlineLedgerRunQuery(`BEGIN TRANSACTION;
     SET TRANSACTION ISOLATION LEVEL REPEATABLE READ;
     SAVEPOINT ${rootSavePointId};
     WITH "CTE" AS (
      SELECT "XaKey" FROM ${databaseNameAsId}.${ledgerTableNameAsId}
        WHERE "EventSubscriber" = ${escapeStr(readModelName)}
        AND "XaKey" = ${escapeStr(xaKey)}
        AND "IsPaused" = FALSE
        AND "Errors" IS NULL
        FOR NO KEY UPDATE NOWAIT
    )
      SELECT 1/Count("CTE"."XaKey") AS "NonZero" FROM "CTE";
    `, true);
  let events = await eventsPromise;

  while (true) {
    if (events.length === 0) {
      throw new PassthroughError();
    }

    let nextCursor = eventstoreAdapter.getNextCursor(cursor, events);
    eventsPromise = eventstoreAdapter.loadEvents({
      eventTypes,
      eventsSizeLimit: 65536000,
      limit: 1000,
      cursor: nextCursor
    }).then(result => result != null ? result.events : []);
    let appliedEventsCount = 0;

    try {
      for (const event of events) {
        const savePointId = generateGuid(xaKey, `${appliedEventsCount}`);

        try {
          const handler = await modelInterop.acquireEventHandler(store, event);

          if (handler != null) {
            await inlineLedgerRunQuery(`SAVEPOINT ${savePointId}`);
            await handler();
            await inlineLedgerRunQuery(`RELEASE SAVEPOINT ${savePointId}`);
            lastSuccessEvent = event;
          }

          appliedEventsCount++;

          if (getVacantTimeInMillis() < 0) {
            nextCursor = eventstoreAdapter.getNextCursor(cursor, events.slice(0, appliedEventsCount));
            localContinue = false;
            break;
          }
        } catch (error) {
          if (error instanceof PassthroughError) {
            throw error;
          }

          nextCursor = eventstoreAdapter.getNextCursor(cursor, events.slice(0, appliedEventsCount));
          await inlineLedgerRunQuery(`ROLLBACK TO SAVEPOINT ${savePointId};
             RELEASE SAVEPOINT ${savePointId}
          `);
          lastFailedEvent = event;
          lastError = error;
          break;
        }
      }
    } catch (originalError) {
      if (originalError instanceof PassthroughError) {
        throw originalError;
      }

      nextCursor = cursor;
      appliedEventsCount = 0;
      const composedError = new Error(`Fatal inline ledger building error: ${originalError.message}`);
      composedError.stack = `${composedError.stack}${originalError.stack}`;
      lastError = composedError;
      lastSuccessEvent = null;
      lastFailedEvent = null;
      await inlineLedgerRunQuery(`ROLLBACK TO SAVEPOINT ${rootSavePointId};
         RELEASE SAVEPOINT ${rootSavePointId}
      `);
    }

    if (lastError == null) {
      await inlineLedgerRunQuery(`UPDATE ${databaseNameAsId}.${ledgerTableNameAsId} SET 
         ${lastSuccessEvent != null ? `"SuccessEvent" = ${escapeStr(JSON.stringify(lastSuccessEvent))},` : ''} 
         "Cursor" = ${escapeStr(JSON.stringify(nextCursor))}
         WHERE "EventSubscriber" = ${escapeStr(readModelName)};

         COMMIT;
        `);
    } else {
      await inlineLedgerRunQuery(`UPDATE ${databaseNameAsId}.${ledgerTableNameAsId}
         SET "Errors" = jsonb_insert(
           COALESCE("Errors", jsonb('[]')),
           CAST(('{' || jsonb_array_length(COALESCE("Errors", jsonb('[]'))) || '}') AS TEXT[]),
           jsonb(${escapeStr(JSON.stringify(serializeError(lastError)))})
         ),
         ${lastFailedEvent != null ? `"FailedEvent" = ${escapeStr(JSON.stringify(lastFailedEvent))},` : ''}
         ${lastSuccessEvent != null ? `"SuccessEvent" = ${escapeStr(JSON.stringify(lastSuccessEvent))},` : ''}
         "Cursor" = ${escapeStr(JSON.stringify(nextCursor))}
         WHERE "EventSubscriber" = ${escapeStr(readModelName)};

         COMMIT;
        `);
    }

    const isBuildSuccess = lastError == null && appliedEventsCount > 0;
    cursor = nextCursor;

    if (getVacantTimeInMillis() < 0) {
      localContinue = false;
    }

    if (isBuildSuccess && localContinue) {
      rootSavePointId = generateGuid(xaKey, 'ROOT');
      await inlineLedgerRunQuery(`BEGIN TRANSACTION;
        SET TRANSACTION ISOLATION LEVEL REPEATABLE READ;
        SAVEPOINT ${rootSavePointId};
        WITH "CTE" AS (
          SELECT "XaKey" FROM ${databaseNameAsId}.${ledgerTableNameAsId}
            WHERE "EventSubscriber" = ${escapeStr(readModelName)}
            AND "XaKey" = ${escapeStr(xaKey)}
            AND "IsPaused" = FALSE
            AND "Errors" IS NULL
            FOR NO KEY UPDATE NOWAIT
        )
          SELECT 1/Count("CTE"."XaKey") AS "NonZero" FROM "CTE";
        `, true);
      events = await eventsPromise;
    } else {
      if (isBuildSuccess) {
        await next();
      }

      throw new PassthroughError();
    }
  }
};

const build = async (basePool, readModelName, store, modelInterop, next, eventstoreAdapter, getVacantTimeInMillis, provideLedger) => {
  const {
    PassthroughError,
    inlineLedgerRunQuery,
    schemaName,
    tablePrefix,
    escapeId,
    escapeStr,
    generateGuid
  } = basePool;

  try {
    basePool.activePassthrough = true;
    const databaseNameAsId = escapeId(schemaName);
    const ledgerTableNameAsId = escapeId(`${tablePrefix}__${schemaName}__LEDGER__`);
    const trxTableNameAsId = escapeId(`${tablePrefix}__${schemaName}__TRX__`);
    const xaKey = generateGuid(`${Date.now()}${Math.random()}${process.pid}`);
    const rows = await inlineLedgerRunQuery(`WITH "MaybeAcquireLock" AS (
         SELECT * FROM ${databaseNameAsId}.${ledgerTableNameAsId}
         WHERE "EventSubscriber" = ${escapeStr(readModelName)}
         AND "IsPaused" = FALSE
         AND "Errors" IS NULL
         FOR NO KEY UPDATE NOWAIT
       ), "CleanTrx" AS (
        DELETE FROM ${databaseNameAsId}.${trxTableNameAsId}
        WHERE "Timestamp" < CAST(extract(epoch from clock_timestamp()) * 1000 AS BIGINT) - 86400000
        RETURNING *
      ), "InsertTrx" AS (
        INSERT INTO ${databaseNameAsId}.${trxTableNameAsId}(
          "Timestamp", "XaKey", "XaValue"
        ) VALUES (
          CAST(extract(epoch from clock_timestamp()) * 1000 AS BIGINT), 
          ${escapeStr(xaKey)},
          CAST(pg_backend_pid() AS VARCHAR(190))
        )
        RETURNING *
      )
      UPDATE ${databaseNameAsId}.${ledgerTableNameAsId}
      SET "XaKey" = ${escapeStr(xaKey)}
      WHERE "EventSubscriber" = ${escapeStr(readModelName)}
      AND CAST(COALESCE((SELECT LEAST(Count("InsertTrx".*), 0) FROM "InsertTrx"), 0) AS BIGINT) = 0
      AND CAST(COALESCE((SELECT LEAST(Count("CleanTrx".*), 0) FROM "CleanTrx"), 0) AS BIGINT) = 0
      AND (SELECT Count("MaybeAcquireLock".*) FROM "MaybeAcquireLock") = 1
      AND "IsPaused" = FALSE
      AND "Errors" IS NULL
      RETURNING *
      `);
    let readModelLedger = rows.length === 1 ? rows[0] : null;

    if (readModelLedger == null || readModelLedger.Errors != null) {
      throw new PassthroughError();
    }

    const eventTypes = readModelLedger.EventTypes != null ? readModelLedger.EventTypes : null;

    if (!Array.isArray(eventTypes) && eventTypes != null) {
      throw new TypeError('eventTypes');
    }

    const cursor = readModelLedger.Cursor != null ? readModelLedger.Cursor : null;

    if (cursor != null && cursor.constructor !== String) {
      throw new TypeError('cursor');
    }

    await provideLedger(readModelLedger);
    const currentPool = {
      ledgerTableNameAsId,
      databaseNameAsId,
      eventTypes,
      inputCursor: cursor,
      readModelLedger,
      xaKey
    };
    const buildMethod = cursor == null ? buildInit : buildEvents;
    await buildMethod(currentPool, basePool, readModelName, store, modelInterop, next, eventstoreAdapter, getVacantTimeInMillis, provideLedger);
  } catch (error) {
    if (!(error instanceof PassthroughError)) {
      throw error;
    }

    try {
      await inlineLedgerRunQuery(`ROLLBACK`);
    } catch (err) {
      if (!(err instanceof PassthroughError)) {
        throw err;
      }
    }
  } finally {
    basePool.activePassthrough = false;
  }
};

var _default = build;
exports.default = _default;
//# sourceMappingURL=build.js.map